import contextlib
from typing import Any, Awaitable, Callable, Iterable, ParamSpec, TypeVar

from .._tonio import CancelledError, ResultHolder, get_runtime
from ._events import Event, Waiter


_T = TypeVar('_T')
_Params = ParamSpec('_Params')
_Return = TypeVar('_Return')


class AsyncSpawnJoin:
    __slots__ = ['_waiter', '_res', '_errs']

    def __init__(self, waiter, res, errs):
        self._waiter = waiter
        self._res = res
        self._errs = errs

    def __await__(self):
        return self._wait().__await__()

    async def _wait(self):
        await self._waiter
        if self._errs:
            raise ExceptionGroup('SpawnExceptionGroup', self._errs)
        return self._res.fetch()


def spawn(*coros) -> Awaitable[Any]:
    events = []
    res = ResultHolder(len(coros))
    errs = []

    async def wrapper(idx, coro, event):
        try:
            ret = await coro
            res.store(ret, idx)
        except Exception as exc:
            errs.append(exc)
        finally:
            event.set()

    for idx, coro in enumerate(coros):
        event = Event()
        events.append(event)
        get_runtime()._spawn_pyasyncgen(wrapper(idx, coro, event))

    waiter = Waiter(*events)
    return AsyncSpawnJoin(waiter, res, errs)


async def spawn_blocking(fn: Callable[_Params, _Return], /, *args: _Params.args, **kwargs: _Params.kwargs) -> _Return:
    ctl, event, res = get_runtime()._spawn_blocking(fn, *args, **kwargs)
    with contextlib.suppress(CancelledError):
        await event.waiter(None)
    err, val = res.fetch()
    if err is None:
        ctl.abort()
        await event.waiter(None)
        err, val = res.fetch()
    if err is True:
        raise val
    return val


def map(fn: Callable[[_T], _Return], /, xs: Iterable[_T]) -> Awaitable[list[_Return]]:
    return spawn(*[fn(x) for x in xs])


def map_blocking(fn: Callable[[_T], _Return], /, xs: Iterable[_T]) -> Awaitable[list[_Return]]:
    return spawn(*[spawn_blocking(fn, x) for x in xs])


async def yield_now():
    event = Event()
    event.set()
    await event.waiter(None)



import contextlib
from typing import Callable, Iterable, ParamSpec, TypeVar

from ._events import Event, Waiter
from ._tonio import CancelledError, ResultHolder, get_runtime
from ._types import Coro


_T = TypeVar('_T')
_Params = ParamSpec('_Params')
_Return = TypeVar('_Return')


def spawn(*coros: Coro):
    events = []
    res = ResultHolder(len(coros))
    errs = []

    def wrapper(idx, coro, event):
        try:
            ret = yield coro
            res.store(ret, idx)
        except Exception as exc:
            errs.append(exc)
        finally:
            event.set()

    for idx, coro in enumerate(coros):
        event = Event()
        events.append(event)
        get_runtime()._spawn_pygen(wrapper(idx, coro, event))

    waiter = Waiter(*events)

    def join():
        yield waiter
        if errs:
            raise ExceptionGroup('SpawnExceptionGroup', errs)
        return res.fetch()

    return join()


def spawn_blocking(fn: Callable[_Params, _Return], /, *args: _Params.args, **kwargs: _Params.kwargs) -> Coro[_Return]:
    ctl, event, res = get_runtime()._spawn_blocking(fn, *args, **kwargs)
    with contextlib.suppress(CancelledError):
        yield event.waiter(None)
    err, val = res.fetch()
    if err is None:
        ctl.abort()
        yield event.waiter(None)
        err, val = res.fetch()
    if err is True:
        raise val
    return val


def map(fn: Callable[[_T], _Return], /, xs: Iterable[_T]) -> Coro[list[_Return]]:
    return spawn(*[fn(x) for x in xs])


def map_blocking(fn: Callable[[_T], _Return], /, xs: Iterable[_T]) -> Coro[list[_Return]]:
    return spawn(*[spawn_blocking(fn, x) for x in xs])



import contextlib
from typing import Awaitable, TypeVar

from .._time import _Interval
from .._tonio import CancelledError, ResultHolder, get_runtime
from ._events import Event


_T = TypeVar('_T')


class Interval(_Interval):
    __slots__ = []

    def tick(self) -> Awaitable[None]:
        timeout = self._poll()
        return Event().waiter(timeout)


def sleep(timeout: int | float) -> Awaitable[None]:
    return Event().wait(timeout)


async def timeout(coro: Awaitable[_T], timeout: int | float) -> tuple[None | _T, bool]:
    done = Event()
    res = ResultHolder()
    errs = []

    async def wrapper():
        try:
            ret = await coro
            res.store(ret)
        except CancelledError:
            pass
        except Exception as exc:
            errs.append(exc)
        finally:
            done.set()

    get_runtime()._spawn_pyasyncgen(wrapper())

    await done.wait(timeout)
    if not done.is_set():
        with contextlib.suppress(CancelledError):
            coro.throw(CancelledError)
        return None, False
    if errs:
        [err] = errs
        raise err
    return res.fetch(), True


def interval(period: int | float, at: int | None = None) -> Interval:
    period = round(max(0, period * 1_000_000))
    at = at or get_runtime()._clock
    return Interval(at, period)



import multiprocessing

from ._colored._events import Event as EventAw
from ._events import Event
from ._tonio import ResultHolder, Runtime as _Runtime, set_runtime as _set_runtime
from ._utils import is_asyncg


class Runtime(_Runtime):
    def run_forever(self):
        try:
            self._run_forever_pre()
            self._run()
        finally:
            self._run_forever_post()

    def _run_forever_pre(self):
        # TODO: signals
        pass

    def _run_forever_post(self):
        # TODO: signals
        self._stopping = False

    def run_pygen_until_complete(self, coro):
        done = Event()
        res = ResultHolder()
        is_exc = False

        def runner():
            nonlocal is_exc
            try:
                ret = yield coro
                res.store(ret)
            except Exception as exc:
                is_exc = True
                res.store(exc)
            finally:
                done.set()

        def watcher():
            yield from done.wait()
            self.stop()

        self._spawn_pygen(watcher())
        self._spawn_pygen(runner())
        self.run_forever()

        ret = res.fetch()
        if is_exc:
            raise ret
        return ret

    def run_pyasyncgen_until_complete(self, coro):
        done = EventAw()
        res = ResultHolder()
        is_exc = False

        async def runner():
            nonlocal is_exc
            try:
                ret = await coro
                res.store(ret)
            except Exception as exc:
                is_exc = True
                res.store(exc)
            finally:
                done.set()

        async def watcher():
            await done.wait()
            self.stop()

        self._spawn_pyasyncgen(watcher())
        self._spawn_pyasyncgen(runner())
        self.run_forever()

        ret = res.fetch()
        if is_exc:
            raise ret
        return ret

    def run_until_complete(self, coro):
        runner = self.run_pyasyncgen_until_complete if is_asyncg(coro) else self.run_pygen_until_complete
        return runner(coro)

    def stop(self):
        self._stopping = True


def new(
    context: bool = False,
    threads: int | None = None,
    blocking_threadpool_size: int = 128,
    blocking_threadpool_idle_ttl: int = 30,
) -> Runtime:
    threads = threads or multiprocessing.cpu_count()
    runtime = Runtime(
        threads=threads,
        threads_blocking=blocking_threadpool_size,
        threads_blocking_timeout=blocking_threadpool_idle_ttl,
        context=context,
    )
    _set_runtime(runtime)
    return runtime


def run(
    coro,
    context: bool = False,
    threads: int | None = None,
    blocking_threadpool_size: int = 128,
    blocking_threadpool_idle_ttl: int = 30,
):
    runtime = new(
        context=context,
        threads=threads,
        blocking_threadpool_size=blocking_threadpool_size,
        blocking_threadpool_idle_ttl=blocking_threadpool_idle_ttl,
    )
    return runtime.run_until_complete(coro)



from ._ctl import map as map, map_blocking as map_blocking, spawn as spawn, spawn_blocking as spawn_blocking
from ._deco import main as main
from ._events import Event as Event, Waiter as Waiter
from ._runtime import Runtime as Runtime, new as runtime, run as run  # noqa: F401
from ._scope import scope as scope
from ._tonio import __version__ as __version__
from .time import sleep as sleep
